import pytest

from undate.converters.calendars import HebrewDateConverter
from undate.converters.calendars.hebrew.transformer import HebrewUndate
from undate.undate import Calendar, Undate
from undate.date import DatePrecision, Date


class TestHebrewDateConverter:
    def test_parse(self):
        # day
        # 26 Tammuz 4816: Tammuz = month 4 (17 July, 1056 Gregorian)
        date_str = "26 Tammuz 4816"
        date = HebrewDateConverter().parse(date_str)
        assert date == HebrewUndate(4816, 4, 26)
        assert date.calendar == Calendar.HEBREW
        assert date.precision == DatePrecision.DAY
        assert date.label == f"{date_str} {HebrewDateConverter.calendar_name}"

        # month
        date_str = "Ṭevet 5362"
        date = HebrewDateConverter().parse(date_str)
        assert date == HebrewUndate(5362, 10)  # Teveth = month 10
        assert date.calendar == Calendar.HEBREW
        assert date.precision == DatePrecision.MONTH
        assert date.label == f"{date_str} {HebrewDateConverter.calendar_name}"

        # year
        date_str = "4932"
        date = HebrewDateConverter().parse(date_str)
        assert date == HebrewUndate(4932)
        assert date.calendar == Calendar.HEBREW
        assert date.precision == DatePrecision.YEAR
        assert date.label == f"{date_str} {HebrewDateConverter.calendar_name}"

    def test_gregorian_earliest_latest(self):
        # earliest/latest should be converted to Gregorian for comparison

        # full date

        # 26 Tammuz 4816: 17 July, 1056; Tammuz = month 4
        date = HebrewUndate(4816, 4, 26)
        assert date.earliest == Date(1056, 7, 17)
        assert date.latest == Date(1056, 7, 17)
        # 13 Tishrei 5416 Anno Mundi (1655-10-14)
        date = HebrewUndate(5416, 7, 13)  # Tishrei = month 7
        assert date.earliest == Date(1655, 10, 14)
        assert date.latest == Date(1655, 10, 14)

        # month

        # Ṭevet 5362 Anno Mundi (25 December, 1601 – 22 January, 1602)
        date = HebrewUndate(5362, 10)
        assert date.earliest == Date(1601, 12, 25)
        assert date.latest == Date(1602, 1, 22)

        # year
        # 5416 : October 1655 to September 1656
        date = HebrewUndate(5416)
        assert date.earliest == Date(1655, 10, 2)
        assert date.latest == Date(1656, 9, 18)

    def test_parse_error(self):
        # a string we can't parse should raise an error
        with pytest.raises(ValueError):
            HebrewDateConverter().parse("January 2, 1991")
        # empty string should also error
        with pytest.raises(ValueError):
            HebrewDateConverter().parse("")

        # non-string input should raise a type error
        with pytest.raises(TypeError):
            HebrewDateConverter().parse(42)

        with pytest.raises(TypeError):
            HebrewDateConverter().parse({"foo": "bar"})

    def test_partially_known(self):
        # hebrew dates get existing partially unknown behavior

        converter = HebrewDateConverter()

        # hebrew first/last month are not the same as min/max
        unknown_month = HebrewUndate(1243, "XX")
        assert unknown_month.precision == DatePrecision.MONTH
        assert unknown_month.earliest == Date(
            *converter.to_gregorian(1243, converter.first_month(), 1)
        )
        last_month = converter.last_month(year=1243)
        assert unknown_month.latest == Date(
            *converter.to_gregorian(
                1243, last_month, converter.max_day(1243, last_month)
            )
        )

        partially_unknown_month = HebrewUndate(1243, "1X")
        assert partially_unknown_month.precision == DatePrecision.MONTH
        assert partially_unknown_month.earliest == Date(
            *converter.to_gregorian(1243, 10, 1)
        )
        # for unknown digit, assume largest possible value instead
        # of last semantic monthin the year
        last_month = converter.max_month(year=1243)
        last_day = converter.max_day(1243, last_month)
        assert partially_unknown_month.latest == Date(
            *converter.to_gregorian(1243, last_month, last_day)
        )

        # second month has 29 days
        unknown_day = HebrewUndate(1243, 2, "XX")
        assert unknown_day.precision == DatePrecision.DAY
        assert unknown_day.earliest == Date(*converter.to_gregorian(1243, 2, 1))
        assert unknown_day.latest == Date(*converter.to_gregorian(1243, 2, 29))

        partially_unknown_day = HebrewUndate(1243, 2, "2X")
        assert partially_unknown_day.precision == DatePrecision.DAY
        assert partially_unknown_day.earliest == Date(
            *converter.to_gregorian(1243, 2, 20)
        )
        assert partially_unknown_day.latest == Date(
            *converter.to_gregorian(1243, 2, 29)
        )

    def test_compare_across_calendars(self):
        # only day-precision dates can be exactly equal across calendars

        # 26 Tammuz 4816: Tammuz = month 4 (17 July, 1056 Gregorian)
        assert HebrewUndate(4816, 4, 26) == Undate(1056, 7, 17)
        # 13 Tishrei 5416; Tieshrei = month 7 (1655-10-14)
        assert HebrewUndate(5416, 7, 13) == Undate(1655, 10, 14)

        # greater than / less than
        assert HebrewUndate(4816) < Undate(1060)
        assert HebrewUndate(5416) < Undate(1660)
        assert HebrewUndate(5416, 7) > Undate(1655, 1)
        assert HebrewUndate(4816, 4, 26) > Undate(1055, 5)

        # 26 Tammuz 4816: Tammuz = month 4 (17 July, 1056)
        # so it falls within or is c ontained by July 1056
        assert HebrewUndate(4816, 4, 26) in Undate(1056, 7)
        assert HebrewUndate(4816, 4, 26) not in Undate(1054)

        # sorting
        sorted_dates = sorted(
            [
                HebrewUndate(4816, 4, 26),  # 1056-07-17
                HebrewUndate(5416),  # 1655
                HebrewUndate(500),  # -3261
                Undate(1995),
                Undate(33),
                Undate(1350),
            ]
        )
        expected_gregorian_years = [-3261, 33, 1056, 1350, 1655, 1995]
        assert [d.earliest.year for d in sorted_dates] == expected_gregorian_years

    def test_days_in_year(self):
        converter = HebrewDateConverter()
        assert converter.days_in_year(4816) == 353
        assert converter.days_in_year(4817) == 355
        assert converter.days_in_year(4818) == 384
        assert converter.days_in_year(4819) == 355

    def test_representative_years(self):
        converter = HebrewDateConverter()
        # single year is not filtered
        assert converter.representative_years([4816]) == [4816]
        # 4816 has 353 days; 4817 has 355; 4818 has 384; 4819 has 355
        assert converter.representative_years([4816, 4817, 4818, 4819]) == [
            4816,
            4817,
            4818,
        ]
        assert converter.representative_years([4816, 4817, 4818, 4819, 4837]) == [
            4816,
            4817,
            4818,
            4837,
        ]

        # if no years are provided, returns a known leap year and non-leap years
        assert converter.representative_years() == [
            converter.LEAP_YEAR,
            converter.NON_LEAP_YEAR,
        ]
        assert converter.representative_years([]) == [
            converter.LEAP_YEAR,
            converter.NON_LEAP_YEAR,
        ]



from undate.converters.calendars import HebrewDateConverter
from undate.undate import Calendar


class SeleucidDateConverter(HebrewDateConverter):
    #: offset for Seleucid calendar: Seleucid year + 3449 = Anno Mundi year
    SELEUCID_OFFSET = 3449

    #: converter name: Seleucid
    name: str = "Seleucid"
    calendar_name: str = "Seleucid"

    def __init__(self):
        super().__init__()
        # override hebrew calendar to initialize undates with seleucid
        # calendar; this triggers Seleucid calendar to_gregorian method use
        self.transformer.calendar = Calendar.SELEUCID

    def to_gregorian(self, year: int, month: int, day: int) -> tuple[int, int, int]:
        """Convert a Seleucid date, specified by year, month, and day,
        to the Gregorian equivalent date. Uses hebrew calendar conversion
        logic with :attr:`SELEUCID_OFFSET`. Returns a tuple of year, month, day.
        """
        return super().to_gregorian(year + self.SELEUCID_OFFSET, month, day)

    def days_in_year(self, year: int) -> int:
        """the number of days in the specified year for this calendar"""
        return super().days_in_year(year + self.SELEUCID_OFFSET)



import pytest
from undate.converters.edtf.parser import edtf_parser

# for now, just test that valid dates can be parsed

testcases = [
    "1984",
    "1984-05",
    "1984-12",
    "1001-03-30",
    "1901-02-20",
    "1000/2000",
    "1000-01/2000-05-01",
    # level 1
    "Y170000002",
    "Y-170000002",
    "2001-21",  # spring 2001
    # negative year
    "-1985",
    # qualifiers
    "1984?",
    "2004-06~",
    "2004-06-11%",
    # unspecified digits from right
    "201X",
    "20XX",
    "2004-XX",
    "1985-04-XX",
    "1985-XX-XX",
    # open ended intervals
    "1985-04-12/..",
    "1985-04/..",
    "../1985-04-12",
    "/1985-04-12",
    # level 2 unspecified digits
    "156X-12-25",
    "XXXX-12-XX",
    "1XXX-12",
    "1XXX-XX",
    "1984-1X",
]


@pytest.mark.parametrize("date_string", testcases)
def test_should_parse(date_string):
    assert edtf_parser.parse(date_string)


error_cases = ["1984-13", "Y1702", "1984-00", "1984-01-00"]


@pytest.mark.parametrize("date_string", error_cases)
def test_should_error(date_string):
    with pytest.raises(Exception):
        edtf_parser.parse(date_string)



from typing import Dict, List, Union

from undate import Undate, UndateInterval
from undate.converters.base import BaseDateConverter


class ISO8601DateFormat(BaseDateConverter):
    # NOTE: do we care about validation? could use regex
    # but maybe be permissive, warn if invalid but we can parse

    #: converter name: ISO8601
    name: str = "ISO8601"
    # do not change; Undate relies on this string

    #: datetime strftime format for known part of date
    iso_format: Dict[str, str] = {
        "year": "%Y",
        "month": "%m",
        "day": "%d",
    }

    def parse(self, value: str) -> Union[Undate, UndateInterval]:
        """
        Parse an ISO88601 string and return an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval`. Currently supports
        YYYY, YYYY-MM, YYYY-MM-DD, --MM-DD for single date
        and interval format (YYYY/YYYY in any supported single date format).
        """
        # TODO: what happens if someone gives us a full isoformat date with time?
        # (ignore, error?)
        # TODO: what about invalid format?
        parts: List[str] = value.split("/")  # split in case we have a range
        if len(parts) == 1:
            return self._parse_single_date(parts[0])
        elif len(parts) == 2:
            # date range; parse both parts and initialize an interval
            start, end = [self._parse_single_date(p) for p in parts]
            return UndateInterval(start, end)
        else:
            # more than two parts = unexpected input
            raise ValueError

    def _parse_single_date(self, value: str) -> Undate:
        # split single iso date into parts; convert to int or None
        # special case: missing year
        date_parts: List[Union[int, None]] = []
        if value.startswith("--"):
            date_parts.append(None)  # year unknown
            value = value[2:]

        date_parts.extend([int(p) if p else None for p in value.split("-")])
        # TODO: handle too many parts. Getting this error from type checking:
        # Argument of type "int | None" cannot be assigned to parameter "formatter" of type "BaseDateFormat | None" in function "__init__"
        return Undate(*date_parts)  # type: ignore

    def to_string(self, undate: Union[Undate, UndateInterval]) -> str:
        """
        Convert an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval` to ISO8601 string format.
        """
        if isinstance(undate, Undate):
            return self._undate_to_string(undate)
        elif isinstance(undate, UndateInterval):
            # strictly speaking I don't think ISO8601 supports open-ended ranges
            # should we add an exception for dates that can't be represented by a particular format?
            # (we'll likely need it for uncertain/approx, which ISO8601 doesn't handle')
            start = self._undate_to_string(undate.earliest) if undate.earliest else ""
            end = self._undate_to_string(undate.latest) if undate.latest else ""
            return f"{start}/{end}"

    def _undate_to_string(self, undate: Undate) -> str:
        # serialize to iso format for simplicity, for now
        date_parts: List[Union[str, None]] = []
        # for each part of the date that is known, generate the string format
        # then combine
        # TODO: should error if we have year and day but no month
        # TODO: may want to refactor and take advantage of the year/month/day properties
        # added for use in EDTF formatter code
        for date_portion, iso_format in self.iso_format.items():
            # is known means fully known, means guaranteed integer
            if undate.is_known(date_portion):
                # NOTE: datetime strftime for %Y for 3-digit year
                # results in leading zero in some environments
                # and not others; force year to always be 4 digits
                if date_portion == "year" and undate.year:
                    try:
                        date_parts.append("%04d" % int(undate.year))
                    except ValueError:
                        # shouldn't happen because of is_known
                        date_parts.append(undate.year)
                elif date_portion == "month" and undate.month:
                    try:
                        date_parts.append("%02d" % int(undate.month))
                    except ValueError:
                        # shouldn't happen because of is_known
                        date_parts.append(undate.month)
                elif date_portion == "day" and undate.day:
                    try:
                        date_parts.append("%02d" % int(undate.day))
                    except ValueError:
                        # shouldn't happen because of is_known
                        date_parts.append(undate.day)

            elif date_portion == "year":
                # if year is not known, add '-' for year portion,
                # to genereate --MM-DD unknown year format
                date_parts.append("-")
        # TODO: fix type error: "list[str | None]" is incompatible with "Iterable[str]"
        return "-".join(date_parts)  # type: ignore
